"""API for `shamo.hd_tdcs`."""
from .simulation.single.problem import ProbHDTDCSSim
from .simulation.single.solution import SolHDTDCSSim

from .simulation.parametric.problem import ProbParamHDTDCSSim
from .simulation.parametric.solution import SolParamHDTDCSSim
from .simulation.parametric.surrogate import (
    SurrMaskedScalarNiiJ,
    SurrMaskedScalarNiiMagJ,
    SurrMaskedScalarNiiV,
)


__all__ = [
    "ProbHDTDCSSim",
    "SolHDTDCSSim",
    "ProbParamHDTDCSSim",
    "SolParamHDTDCSSim",
    "SurrMaskedScalarNiiJ",
    "SurrMaskedScalarNiiMagJ",
    "SurrMaskedScalarNiiV",
]



"""Implement the `PointSensor` class."""
from collections.abc import Iterable

from shamo.core.fem import SensorABC, Group


class PointSensor(SensorABC):
    """A FEM sensor.

    Parameters
    ----------
    tissue : str
        The tissue the sensor is in.
    point : shamo.fem.Group
        The physical group of the sensor.
    node : int
        The tag of the node corresponding to the sensor.
    """

    def __init__(self, tissue, real_coords, mesh_coords, point, node, **kwargs):
        super().__init__(tissue, SensorABC.TYPE_POINT, real_coords, mesh_coords)
        self.update({"point": Group(**point), "node": int(node)})

    @property
    def point(self):
        """Return the group of the sensor.

        Returns
        -------
        shamo.fem.Group
            The group of the sensor.
        """
        return self["point"]

    @property
    def node(self):
        """Return the node tag of the sensor.

        Returns
        -------
        int
            The node tag of the sensor.
        """
        return self["node"]



"""API for `shamo.core.surrogate`."""
from .abc import SurrABC
from .scalar import SurrScalar
from .masked_scalar import SurrMaskedScalar, SurrMaskedScalarNii


__all__ = ["SurrABC", "SurrScalar", "SurrMaskedScalar", "SurrMaskedScalarNii"]



"""Implement the `Sensor` class."""


class SensorABC(dict):
    """A FEM sensor.

    Parameters
    ----------
    tissue : str
        The tissue the sensor is in.
    real_coords : Iterable [float]
        The coordinates of the sensor in the real world.
    mesh_coords : Iterable [float]
        The coordinates of the sensor in the mesh.
    sensor_type : str
        The type of the sensor.
    """

    TYPE_POINT = "point"
    TYPE_CIRCLE = "circle"

    def __init__(self, tissue, sensor_type, real_coords, mesh_coords):
        super().__init__(
            {
                "tissue": tissue,
                "sensor_type": sensor_type,
                "real_coords": tuple(real_coords),
                "mesh_coords": tuple(mesh_coords),
            }
        )

    @property
    def real_coords(self):
        """Return the real coordinates of the sensor.

        Returns
        -------
        tuple (float, float, float)
            The real coordinates of the sensor [m].
        """
        return self["real_coords"]

    @property
    def mesh_coords(self):
        """Return the mesh coordinates of the sensor.

        Returns
        -------
        tuple (float, float, float)
            The mesh coordinates of the sensor [m].
        """
        return self["mesh_coords"]

    @property
    def tissue(self):
        """Return the tissue the sensor is in.

        Returns
        -------
        str
            The tissue the sensor is in.
        """
        return self["tissue"]

    @property
    def sensor_type(self):
        """Return the type of the sensor.

        Returns
        -------
        str
            The type of the sensor.
        """
        return self["sensor_type"]

    @staticmethod
    def load(sensor_type, **kwargs):
        from .point import PointSensor
        from .surface import CircleSensor

        sensor_types = {
            SensorABC.TYPE_POINT: PointSensor,
            SensorABC.TYPE_CIRCLE: CircleSensor,
        }
        return sensor_types[sensor_type](**kwargs)



"""Implement the `SurfSensorABC` and `CircleSensor` classes."""
from shamo.core.fem import SensorABC, Group


class SurfSensorABC(SensorABC):
    """The base class for any surfacic sensor.

    Parameters
    ----------
    tissue : str
        The tissue the sensor is in.
    real_coords : Iterable [float]
        The coordinates of the sensor in the real world.
    mesh_coords : Iterable [float]
        The coordinates of the sensor in the mesh.
    surf : shamo.fem.Group
        The physical group of the sensor.
    """

    def __init__(self, tissue, sensor_type, real_coords, mesh_coords, surf, **kwargs):
        super().__init__(tissue, sensor_type, real_coords, mesh_coords)
        self.update(
            {
                "real_coords": tuple(real_coords),
                "mesh_coords": tuple(mesh_coords),
                "surf": Group(**surf),
            }
        )

    @property
    def surf(self):
        """Return the group of the sensor.

        Returns
        -------
        shamo.fem.Group
            The group of the sensor.
        """
        return self["surf"]


class CircleSensor(SurfSensorABC):
    """The base class for any surfacic sensor.

    Parameters
    ----------
    tissue : str
        The tissue the sensor is in.
    real_coords : Iterable [float]
        The coordinates of the sensor in the real world.
    mesh_coords : Iterable [float]
        The coordinates of the sensor in the mesh.
    surf : shamo.fem.Group
        The physical group of the sensor.
    radius : float
        The radius of the sensor.
    """

    def __init__(self, tissue, real_coords, mesh_coords, surf, radius, **kwargs):
        super().__init__(tissue, SensorABC.TYPE_CIRCLE, real_coords, mesh_coords, surf)
        self.update({"radius": float(radius)})

    @property
    def radius(self):
        """Return the radius of the sensor.

        Returns
        -------
        float
            The radius of the sensor.
        """
        return self["radius"]



"""API for `shamo.core.fem`."""
from .group import Group
from .field import Field
from .tissue import Tissue
from .sensors.abc import SensorABC
from .sensors.point import PointSensor
from .sensors.surface import SurfSensorABC, CircleSensor

from .fem import FEM

__all__ = [
    "FEM",
    "Field",
    "Group",
    "Tissue",
    "SensorABC",
    "PointSensor",
    "SurfSensorABC",
    "CircleSensor",
]



from abc import abstractclassmethod

import nibabel as nib
import numpy as np

from shamo.core.surrogate import SurrScalar


class SurrMaskedScalar(SurrScalar):
    """Generate a Gaussian process from a set of scalar training data.

    Parameters
    ----------
    name : str
        The name of the generated surrogate model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the surrogate model.

    Other Parameters
    ----------------
    params : list [tuple [str, shamo.DistABC]]
        A list of tuples containing the names of the parameters and the
        corresponding distributions as values.
    sol_json_path : str
        The path to the parametric solution the surrogate is built of.
    is_sobol_available : bool
        If ``True``, Sobol indices are already available.
    suffix : str
        The suffix appended to the queried files.

    See Also
    --------
    shamo.core.surrogate.SurrScalar
    """

    def __init__(self, name, parent_path, **kwargs):
        super().__init__(name, parent_path, **kwargs)
        self.update({"suffix": kwargs.get("suffix", "")})

    @classmethod
    def fit(cls, name, parent_path, sol, **kwargs):
        """Generate a Gaussian process from a set of training data.

        Parameters
        ----------
        name : str
            The name of the generated surrogate model.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the surrogate model.
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.

        Returns
        -------
        shamo.Surrogate
            The generated surrogate model.

        Other Parameters
        ----------------
        kernel : sklearn.gaussian_process.kernels.Kernel
            The kernel used to generate the Gaussian process.
        n_restarts_optimizer : int
            The number of restarts for the optimisation step.
        random_state : int
            The seed for the random state.
        alpha : float
            The added diagonal to account for noise on training points.
        n_proc : int
            The number of jobs to run in parallel. ``None`` means 1 job runs at a time
            and ``-1`` means all cores are used.
        metric : func
            The metric to compute.
        suffix : str
            The suffix to append to the file query.

        See Also
        --------
        sklearn.gaussian_process.GaussianProcessRegressor
        """
        surr = super().fit(name, parent_path, sol, **kwargs)
        surr["suffix"] = kwargs.get("suffix", "")
        surr.save()
        return surr

    @classmethod
    def _get_data(cls, sol, suffix="", metric=None, mask=None, **kwargs):
        """Extract relevant data from a parametric solution.

        Parameters
        ----------
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.

        Returns
        -------
        numpy.ndarray
            The coordinates of the training points in the parameter space. Each row
            represents an evaluation.
        numpy.ndarray
            The observations of the actual model at each coordinate from `x`. Each row
            represents an observation.
        list [tuple [str, shamo.DistABC]]
            A list of tuples containing the names of the parameters and the
            corresponding distributions as values.

        Other Parameters
        ----------------
        suffix : str
            The suffix to append to the queried files.
        metric : func
            A function to run on the masked data. It must take one parameter.
        mask : numpy.ndarray
            The mask to apply to the data.

        See Also
        --------
        shamo.core.surrogate.SurrABC
        """
        x = []
        y = []
        for s in sol.get_sub_sols():
            x.append(sol.get_x(s))
            y.append(metric(cls._get_masked_data(s, suffix, mask)))
        x = np.array(x).reshape((len(x), -1))
        y = np.array(y).reshape((len(y),))
        return x, y, sol.get_params()

    @abstractclassmethod
    def _get_masked_data(cls, sol, suffix, mask):
        """Return a linear array containing the masked data.

        Parameters
        ----------
        sol : shamo.core.solutions.single.SolABC
            A solution.
        suffix : str
            The suffix appended to the filename without the extension.
        mask : numpy.ndarray
            A boolean numpy array used to mask the data.

        Returns
        -------
        numpy.ndarray
            The masked data.
        """


class SurrMaskedScalarNii(SurrMaskedScalar):
    """Generate a Gaussian process from a set of scalar training data.

    Parameters
    ----------
    name : str
        The name of the generated surrogate model.
    parent_path : str, byte or os.PathLike
        The path to the parent directory of the surrogate model.

    Other Parameters
    ----------------
    params : list [tuple [str, shamo.DistABC]]
        A list of tuples containing the names of the parameters and the
        corresponding distributions as values.
    sol_json_path : str
        The path to the parametric solution the surrogate is built of.
    is_sobol_available : bool
        If ``True``, Sobol indices are already available.
    suffix : str
        The suffix appended to the queried files.

    See Also
    --------
    shamo.core.surrogate.SurrMaskedScalar
    """

    @classmethod
    def fit(cls, name, parent_path, sol, **kwargs):
        """Generate a Gaussian process from a set of training data.

        Parameters
        ----------
        name : str
            The name of the generated surrogate model.
        parent_path : str, byte or os.PathLike
            The path to the parent directory of the surrogate model.
        sol : shamo.core.solutions.parametric.SolParamABC
            The parametric solution to generate a surrogate model for.

        Returns
        -------
        shamo.Surrogate
            The generated surrogate model.

        Other Parameters
        ----------------
        kernel : sklearn.gaussian_process.kernels.Kernel
            The kernel used to generate the Gaussian process.
        n_restarts_optimizer : int
            The number of restarts for the optimisation step.
        random_state : int
            The seed for the random state.
        alpha : float
            The added diagonal to account for noise on training points.
        n_proc : int
            The number of jobs to run in parallel. ``None`` means 1 job runs at a time
            and ``-1`` means all cores are used.
        metric : func
            The metric to compute.
        suffix : str
            The suffix to append to the file query.
        mask : nibabel.Nifti1Image
            A NIFTI image to use as a mask.

        See Also
        --------
        sklearn.gaussian_process.GaussianProcessRegressor
        """
        mask = kwargs.pop("mask")
        data = mask.get_fdata().astype(bool)
        surr = super().fit(name, parent_path, sol, mask=data, **kwargs)
        mask.to_filename(surr.path / f"{surr.name}_mask.nii")
        return surr

    @classmethod
    def _get_masked_data(cls, sol, suffix, mask):
        """Return a linear array containing the masked data.

        Parameters
        ----------
        sol : shamo.core.solutions.single.SolABC
            A solution.
        suffix : str
            The suffix appended to the filename without the extension.
        mask : numpy.ndarray
            A boolean numpy array used to mask the data.

        Returns
        -------
        numpy.ndarray
            The masked data.
        """
        img = nib.load(sol.path / f"{sol.name}_{suffix}.nii")
        return img.get_fdata()[mask]



"""API for `shamo.utils.logging`."""
from contextlib import contextmanager
import logging
import re
import sys

from wurlitzer import sys_pipes


class StreamToLogger(object):
    """Pipe a stream to a logger.

    The implementation is based on the blog post available at:
    http://techies-world.com/how-to-redirect-stdout-and-stderr-to-a-logger-in-python/

    Parameters
    ----------
    logger : logging.Logger
        The logger to pipe the stream to.
    log_level : int
        The logging level of the piped messages.
    pattern : str
        A pattern containing a 'level' and a 'text' group.
    """

    pattern = re.compile("^(?P<level>[\w]*) +: (?P<text>.*)$")

    def __init__(self, logger, log_level=logging.INFO, pattern=""):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""
        self.pattern = re.compile(pattern)

    def write(self, buf):
        """Write a message to the logger.

        Parameters
        ----------
        buf : bytes
            The buffer to read from.
        """
        for line in buf.rstrip().splitlines():
            _log_with_pattern(line, self.logger, self.log_level, self.pattern)

    def flush(self):
        """A dummy flush function."""
        pass


def _log_with_pattern(line, logger, log_level, pattern):
    """Log a text and check a pattern.

    Parameters
    ----------
    line : str
        The line to log.
    logger : logging.Logger
        The logger.
    log_level : int
        The logging level.
    pattern : str
        A pattern containing a 'level' and a 'text' group.
    """
    line = line.strip()
    lines = line.split("\r")
    for line in lines:
        match = None
        if pattern != "":
            match = re.search(pattern, line)
        if match is not None:
            if match.group("level"):
                logger.log(log_level, match.group("text"))
            else:
                logger.log(
                    log_level,
                    f"{match.group('text').strip()} ({match.group('percentage')})",
                )
        else:
            if line.strip() != "":
                logger.log(log_level, line)


@contextmanager
def stream_to_logger(logger=None, log_level=logging.INFO, pattern=""):
    """A context manager where everything pushed to stdout is piped to the logger.

    Parameters
    ----------
    logger : logging.Logger
        The logger to pipe the stream to.
    log_level : int
        The logging level of the piped messages.
    pattern : str
        A pattern containing a 'level' and a 'text' group.
    """
    logger = logging.getLogger(__name__) if logger is None else logger
    stream = StreamToLogger(logger, log_level, pattern)
    tmp = sys.stdout
    sys.stdout = stream
    with sys_pipes():
        try:
            yield
        finally:
            sys.stdout = tmp


def subprocess_to_logger(process, logger=None, log_level=logging.INFO, pattern=""):
    """A context manager to pipe the output of a subprocess to the logger.

    Parameters
    ----------
    logger : logging.Logger
        The logger to pipe the stream to.
    log_level : int
        The logging level of the piped messages.
    pattern : str
        A pattern containing a 'level' and a 'text' group.
    """
    with process.stdout as pipe:
        with stream_to_logger(logger, log_level, pattern):
            for line in iter(pipe.readline, b""):
                _log_with_pattern(line.decode("utf-8"), logger, log_level, pattern)
    return process.wait()



from shamo.core.surrogate import SurrMaskedScalarNii


class SurrMaskedScalarNiiJ(SurrMaskedScalarNii):
    @classmethod
    def fit(cls, name, parent_path, sol, **kwargs):
        return super().fit(name, parent_path, sol, suffix="j", **kwargs)


class SurrMaskedScalarNiiMagJ(SurrMaskedScalarNii):
    @classmethod
    def fit(cls, name, parent_path, sol, **kwargs):
        return super().fit(name, parent_path, sol, suffix="mag_j", **kwargs)


class SurrMaskedScalarNiiV(SurrMaskedScalarNii):
    @classmethod
    def fit(cls, name, parent_path, sol, **kwargs):
        return super().fit(name, parent_path, sol, suffix="v", **kwargs)
